// Commodore PET-2001 Clone
// PS/2 keyboard and RS-232 interface
// "Serial Communicator" (U2) firmware
// Glen Kleinschmidt
// www.glensstuff.com
// Nov. 2018
// V1.0

// At the time of writing (9.07 PM 9/Nov/18) this is the highest level of development of this firmware.
// Note that I'm a hardware guy, not(!) a programmer and this code thus far was written with simplicity and
// expediency as the overruling factor (too many projects, not enuf time!). However all required
// functionality is implemented. The lack of finesse is in the primitive routines for receiving and
// writing data to and from the PS/2 keyboard with the current lack of parity checking. However, thus far,
// having spent many hours typing data into my PET and having literally typed thousands and thousands of 
// characters without a single evident communications glitch, motivation to develop the code further at this 
// time is not great. This is not a life-support or mission-critical application! If on the 50,000th key 
// press some glitch due to externally generated EMI causes a T to be received as a #, your lunar lander will
// not crash and kill everyone on board for real. Unless of course you were using my PET clone hardware to
// control your actual real-life lunar lander, in which case that is a rather stupid decision on your part 
// and not my fault.

#include <16f874A.h>
#use delay(clock=16000000)
#use fast_io(B)
#use standard_io(D)
#use standard_io(A)
#use RS232(Baud=38400, Xmit=PIN_C6, RCV=PIN_C7)
#FUSES HS,NOWDT,PUT,NOPROTECT,NOLVP,NOCPD,NOWRT,BROWNOUT

int1     command_parity =  0;
int      command        =  0;
int1     shift          =  0;
int1     extended       =  0;
int      key_code       =  0;
int1     start_bit      =  0;
int1     parity_bit     =  0;   
int1     stop_bit       =  0;
                            
void key_code_to_port()
{

   // This routine communicates the currently keyed character code to the matrix encoder (U1)
  
   output_high(pin_A2);          // Busy line on
   
   delay_us(100);
   
   output_D(key_code);
   
   if(shift==1)
   output_high(pin_A1);
   else
   output_low(pin_A1);
  
   if(extended==1)
   output_high(pin_A0);
   else
   output_low(pin_A0);
   
   output_low(pin_A2);           // Busy line off  
   
}

void serial_read()
{

   // This routine receives data (key make and break codes) from the PS/2 keyboard 
   // B2 = Keyboard serial clock input
   // B1 = Keyboard serial data input
   
   key_code=0;
   start_bit=0;
   parity_bit=0;
   stop_bit=0;
  
   while(input(pin_B2));         // Start bit      
   
   if (input(pin_B1))
   start_bit=1;
     
   while(!input(pin_B2));
       
   while(input(pin_B2));         // Bit 0
      
   if (input(pin_B1))
   bit_set(key_code,0);
   
   while(!input(pin_B2));
        
   while(input(pin_B2));         // Bit 1
   
   if (input(pin_B1))
   bit_set(key_code,1);
      
   while(!input(pin_B2));
        
   while(input(pin_B2));         // Bit 2
  
   if (input(pin_B1))
   bit_set(key_code,2);
   
   while(!input(pin_B2));
         
   while(input(pin_B2));         // Bit 3
   
   if (input(pin_B1))
   bit_set(key_code,3);   
   
   while(!input(pin_B2));
         
   while(input(pin_B2));         // Bit 4
   
   if (input(pin_B1))
   bit_set(key_code,4);  
   
   while(!input(pin_B2));
       
   while(input(pin_B2));         // Bit 5
   
   if (input(pin_B1))
   bit_set(key_code,5);  
   
   while(!input(pin_B2));
       
   while(input(pin_B2));         // Bit 6
   
   if (input(pin_B1))
   bit_set(key_code,6);
         
   while(!input(pin_B2));
     
   while(input(pin_B2));         // Bit 7
   
   if (input(pin_B1))
   bit_set(key_code,7);
   
   while(!input(pin_B2));
         
   while(input(pin_B2));         // Parity bit
   
   if (input(pin_B1))
   parity_bit=1;
   
   while(!input(pin_B2));
        
   while(input(pin_B2));         // Stop bit
   
   if (input(pin_B1))
   stop_bit=1;
   
   while(!input(pin_B2));
      
   printf("%X\n\r", key_code);   // Immediately ping the received PS/2 code out of the RS-232 port
   
}

void serial_write()
{

   // This routine sends data to the PS/2 keyboard 
   // B2 = Keyboard serial clock input
   // B1 = Keyboard serial data input

   set_tris_B(251);              // B2 = output
   output_low(pin_B2);           // Clock line low
   delay_us(100);
   set_tris_B(249);              // B1 and B2 = output
   output_low(pin_B1);           // Data line low
   set_tris_B(253);              // B2 = input (release clock)   
   delay_us(10);
   
   while(input(pin_B2));
   
   output_bit(pin_B1, bit_test(command,0));

   while(!input(pin_B2));
  
   while(input(pin_B2));
   
   output_bit(pin_B1, bit_test(command,1));

   while(!input(pin_B2));
    
   while(input(pin_B2));
   
   output_bit(pin_B1, bit_test(command,2));

   while(!input(pin_B2));
     
   while(input(pin_B2));
   
   output_bit(pin_B1, bit_test(command,3));

   while(!input(pin_B2));
       
   while(input(pin_B2));
   
   output_bit(pin_B1, bit_test(command,4));

   while(!input(pin_B2));
     
   while(input(pin_B2));
   
   output_bit(pin_B1, bit_test(command,5));

   while(!input(pin_B2));
     
   while(input(pin_B2));
   
   output_bit(pin_B1, bit_test(command,6));

   while(!input(pin_B2));
     
   while(input(pin_B2));
   
   output_bit(pin_B1, bit_test(command,7));

   while(!input(pin_B2));

   while(input(pin_B2));                        // Parity bit
   
   output_bit(pin_B1, command_parity);

   while(!input(pin_B2));
  
   while(input(pin_B2));
   
   set_tris_B(255);                             // Release the data line
}

void initialise_keyboard()
{      

   // This routing initialises the PS/2 keyboard with fixed constants to ensure correct operation. The keyboard 
   // must be plugged in at power-on.
   
   command           =  255;                    // Reset keyboard
   command_parity    =  1; 
   serial_write();
         
   delay_ms(1000);
   
   command           =  243;                    // Set typmatic delay
   command_parity    =  1;
   serial_write();
   
   delay_ms(100);
   
   command           =  127;                    // 1 - 2cps 
   command_parity    =  0;
   serial_write();
   
   delay_ms(100);

}

void ps2()
{

   // This routine acts upon data received from the PS/2 Keyboard. PS/2 keyboard operation is enabled when the
   // mode control pin (pin PB0) is switched high. All of the PETs control keys, standard characters and shifted
   // graphics characters are supported. See the associated doccumentation for a printable graphical keyboard 
   // chart/key-map to aid the typist.

   Serial_read();          
   
   if(key_code==0x12)                        // Detect left shift key make code
   shift=1;      
   
   if(key_code==0x59)                        // Detect right shift key make code
   shift=1;
     
   if(key_code==0xF0)                        // Detect break code 
   {                          
      serial_read();          
      
      if(key_code==0x12)                     // Detect release of left shift key
      shift=0;         
                  
      if(key_code==0x59)                     // Detect release of right shift key
      shift=0;
      
      key_code=0;                            // Set an invalid key code to release key    
   }
     
   if(key_code==0xE0)                        // Detect Extended key make code
   {     
      serial_read();                         // Read extended key 
      
      if(key_code==0x12 || key_code==0x59)   // Detect and exclude extended status for extended shift
      extended=0;
      else
      extended=1;                            // Extended = true

      if(key_code==0xF0)                     // Detect Extended key release code
      {
         serial_read();
         
         extended=0;
         
         key_code=0;                         // Set an invalid key code to release key
      }
   }
   
   key_code_to_port();                       // Send key_code to matrix encoder

}

void rs232()
{

   // This routine permits communication with the PET via an RS-232 terminal by receiving standard ASCII character
   // and control codes via the RS-232 port instead of communicating with a PS/2 keyboard. This mode is active when
   // the mode control pin (pin PB0) is switched low. Received ASCII codes are immediately echoed back to the terminal.
   // With the exception of the "pi" character, the PET's shifted graphics characters are otherwise not supported in 
   // this terminal operating mode. 
   
   // Received char.    ASCII code(dec.)     PET_key/function
   //
   // H                 72                   Home cursor
   // W                 87                   Cursor up
   // Z                 90                   Cursor down
   // A                 65                   Cursor left
   // S                 83                   Cursor right
   // C                 67                   Clear screen
   // B                 66                   Stop
   // N                 78                   Run
   // R                 82                   Reverse on
   // T                 84                   Reverse off
   // I                 73                   Insert
   // a                 97                   A
   // b                 98                   B
   // c                 99                   C
   // d                 100                  D
   // e                 101                  E
   // f                 102                  F
   // g                 103                  G
   // h                 104                  H
   // i                 105                  I
   // j                 106                  J
   // k                 107                  K
   // l                 108                  L
   // m                 109                  M
   // n                 110                  N
   // o                 111                  O
   // p                 112                  P
   // q                 113                  Q
   // r                 114                  R
   // s                 115                  S
   // t                 116                  T
   // u                 117                  U
   // v                 118                  V
   // w                 119                  W
   // x                 120                  X
   // y                 121                  Y
   // z                 122                  Z
   // 0                 48                   0
   // 1                 49                   1
   // 2                 50                   2
   // 3                 51                   3
   // 4                 52                   4
   // 5                 53                   5
   // 6                 54                   6
   // 7                 55                   7
   // 8                 56                   8
   // 9                 57                   9
   // +                 43                   +
   // -                 45                   -
   // =                 61                   =
   // [                 91                   [
   // ]                 93                   ]
   // <                 60                   <
   // >                 62                   >
   // !                 33                   !
   // @                 64                   @
   // #                 35                   #
   // $                 36                   $
   // %                 37                   %
   // ^                 94                   ^
   // &                 38                   &
   // *                 42                   *
   // (                 40                   (
   // )                 41                   )
   // _                 95                   Left arrow symbol
   // P                 80                   pi symbol
   // \                 92                   \
   // /                 47                   /
   // ;                 59                   ;
   // :                 58                   :
   // '                 39                   '
   // "                 34                   "
   // ,                 44                   ,
   // .                 46                   .
   // Enter             13                   Return
   // Backspace         8                    Delete
  
   key_code=getch();                         // Receive ASCII character from terminal
   printf("%u\n\r", key_code);               // Ping received character back to terminal 
    
   int const ascii_convert_ps2 [128]   =     {  0,114, 20,102,  0,  0,  0,  0,102,  0,
                                                0,  0,  0, 90,108,  0,  0,114,  0,  0,
                                                0,  0, 20,108,116,  0,116,  0,  0,  0,
                                                0,  0, 41, 22, 14, 38, 37, 46, 61, 13,
                                               70, 69,124,121, 82,123,113, 74,112,105,
                                              114,122,107,115,116,108,117,125, 62, 76,
                                               65, 85, 73, 74, 30,116, 17,108,  0,  0,
                                                0,  0,108,102,  0,  0,  0,  0, 17,  0,
                                               54,  0, 20,116, 20,  0,  0,114,  0,  0,
                                              114, 84, 93, 91, 54, 78,  0, 28, 50, 33,
                                               35, 36, 43, 52, 51, 67, 59, 66, 75, 58,
                                               49, 68, 77, 21, 45, 27, 44, 60, 42, 29,
                                               34, 53, 26, 84, 93, 91,  0,  0};  
                                                                                                                                                                                   
   int const ascii_convert_ps2_shift [128] = {  0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,1,0,1,0,0,
                                                0,0,0,1,0,0,0,0,1,0,
                                                1,0,0,0,1,0,0,1,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0};
                                                          
int const ascii_convert_ps2_extended [128]=  {  0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,1,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,1,0,1,0,0,
                                                0,0,1,0,0,0,0,0,0,0,
                                                0,0,0,1,0,0,0,1,0,0,
                                                1,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0,0,0,
                                                0,0,0,0,0,0,0,0};                                                                                                            
  
   shift    =  ascii_convert_ps2_shift[key_code];
   extended =  ascii_convert_ps2_extended[key_code];
   key_code =  ascii_convert_ps2[key_code];
     
   key_code_to_port();         // Send key_code to matrix encoder
   key_code=0;                 // Set an invalid key code to release key
   delay_ms(40);               // Wait for at least one full keyboard scan interval
   key_code_to_port();         // Send key_code to matrix encoder

}

void MAIN()
{

   set_tris_B(255); 
   delay_ms(500);
   initialise_keyboard();
   
   loop:

      if(input(pin_B0))
      {
         ps2();
      }
      else
      {
         rs232();
      }
     
   goto loop;
     
}

